import subprocess
import ujson
import openpost

from troi.operations import is_homogeneous
import troi


class PlaylistElement(troi.Element):

    def __init__(self):
        self.playlist = None
        self.entities = None

    @staticmethod
    def inputs(self):
        return [Recording]

    @property
    def recordings(self):
        return self.entities

    def read(self, inputs, debug=False):

        entities = inputs[0]
        if not is_homogeneous(entities):
            raise TypeError("entity list not homogeneous")

        listens = []
        for e in entities:
            artist_mbids = [ str(mbid) for mbid in e.artist.mbids or [] ]
            listens.append({
                'listened_at' : 0,
                'track_metadata' : {
                    'artist_name' : e.artist.name,
                    'track_name' : e.name,
                    'release_name' : e.release.name if e.release else "",
                    'additional_info' : {
                        'recording_mbid' : str(e.mbid),
                        'artist_mbids' : artist_mbids
                    }
                }
            })


        self.entities = entities
        self.playlist = ujson.dumps(listens, indent=4, sort_keys=True)

        return entities


    def print(self):

        if not self.playlist:
            print("[no playlist generated yet]")
            return

        for recording in self.entities:
            print("%-60s %-50s" % (recording.name[:59], recording.artist.name[:49]))


    def launch(self):

        if not self.playlist:
            raise RuntimeError("Playlist has not been generated yet.")

        op = openpost.OpenPost('https://listenbrainz.org/player', keep_file=True, file_name="playlist.html")
        op.add_key('listens', self.playlist)
        op.send_post()




import requests
import ujson
from troi import Element, Artist, Release, Recording
import pylistenbrainz


class UserArtistsElement(Element):
    '''
        Fetch artist statistics from ListenBrainz
    '''

    def __init__(self, user_name, count=25, offset=0, time_range='all_time'):
        self.client = pylistenbrainz.ListenBrainz()
        self.user_name = user_name
        self.count = count
        self.offset = offset
        self.time_range = time_range

    def outputs(self):
        return [Artist]

    def read(self, inputs = [], debug=False):

        artist_list = []
        artists = self.client.get_user_artists(self.user_name, self.count, self.offset, self.time_range)
        for a in artists['payload']['artists']:
            artist_list.append(Artist(a['artist_name'], mbids=a['artist_mbids'], msid=a['artist_msid']))

        return artist_list


class UserReleasesElement(Element):
    '''
        Fetch release statistics from ListenBrainz
    '''

    def __init__(self, user_name, count=25, offset=0, time_range='all_time'):
        self.client = pylistenbrainz.ListenBrainz()
        self.user_name = user_name
        self.count = count
        self.offset = offset
        self.time_range = time_range

    def outputs(self):
        return [Release]

    def read(self, inputs = [], debug=False):

        release_list = []
        releases = self.client.get_user_releases(self.user_name, self.count, self.offset, self.time_range)
        for r in releases['payload']['releases']:
            artist = Artist(r['artist_name'], mbids=r['artist_mbids'], msid=r['artist_msid'])
            release_list.append(Release(r['release_name'], mbid=r['release_mbid'], msid=r['release_msid'], 
                                artist=artist))

        return release_list


class UserRecordingElement(Element):
    '''
        Fetch recording statistics from ListenBrainz
    '''

    def __init__(self, user_name, count=25, offset=0, time_range='all_time'):
        self.client = pylistenbrainz.ListenBrainz()
        self.user_name = user_name
        self.count = count
        self.offset = offset
        self.time_range = time_range

    def outputs(self):
        return [Recording]

    def read(self, inputs = [], debug=False):
        recording_list = []
        recordings = self.client.get_user_recordings(self.user_name, self.count, self.offset, self.time_range)
        for r in recordings['payload']['recordings']:
            artist = Artist(r['artist_name'], mbids=r['artist_mbids'], msid=r['artist_msid'])
            release = Release(r['release_name'], mbid=r['release_mbid'], msid=r['release_msid'])
            recording_list.append(Recording(r['track_name'], mbid=r['recording_mbid'], msid=r['recording_msid'], 
                                  artist=artist, release=release))

        return recording_list



from collections import defaultdict
import copy
import json
import os
import sys
import traceback
import requests.exceptions

from flask import Flask, render_template, request, jsonify
from werkzeug.exceptions import NotFound, BadRequest, InternalServerError, \
                                MethodNotAllowed, ImATeapot, ServiceUnavailable
import troi.utils
import troi.playlist

TEMPLATE_FOLDER = os.path.join(os.path.dirname(os.path.realpath(__file__)), "template")
PATCH_FOLDER = "/app/patches"

sys.stderr.write(TEMPLATE_FOLDER + "\n")

registered_queries = {}
app = Flask(__name__,
            template_folder = TEMPLATE_FOLDER)

@app.route('/')
def index():
    """ The home page that shows all of the available queries."""
    return render_template("index.html", patches=patches)


@app.errorhandler(404)
def page_not_found(e):
    return render_template('error.html', error="Patch not found."), 404


def error_check_arguments(inputs):

    args = []
    for input in inputs:
        arg = request.args.get(input['name'], '')
        if not arg and not input['optional']:
            return "Parameter %s is missing." % input['name'], ()
        try:
            args.append(input['type'](arg))
        except:
            if input['optional']:
                args.append(None)
            else:
                return "Parameter %s is of incorrect type. Must be %s." % (input['name'], input['type']), ()
                

    return "", args


def web_patch_handler():
    """
        This is the view handler for the web page.
    """
  
    patch_name = request.path[1:]
    try:
        patch = patches[patch_name]()
    except KeyError:
        return render_template("error.html", error="Cannot load patch %s" % patch_name), 404

    inputs = patch.inputs()
    outputs = patch.outputs()
    desc = patch.description()

    recordings = []
    post_data = ""
    error = ""
    args = []
    if len(request.args):
        error, args = error_check_arguments(inputs)
        if not error:
            try:
                pipeline = patch.create(args)
            except (BadRequest, InternalServerError, ImATeapot, ServiceUnavailable, NotFound, RuntimeError, requests.exceptions.HTTPError) as err:
                error = err
            except RuntimeError as err:
                error = err
            except Exception as err:
                error = traceback.format_exc()
                print(error)

            if not error:
                try:
                    playlist = troi.playlist.PlaylistElement()
                    playlist.set_sources(pipeline)
                    playlist.generate()
                    recordings = playlist.recordings
                    post_data = playlist.playlist
                except (RuntimeError, requests.exceptions.HTTPError) as err:
                    error = str(err)

    return render_template("patch.html",
                           error=error,
                           recordings=recordings,
                           count=len(recordings) if recordings else -1,
                           inputs=inputs,
                           columns=outputs,
                           args=args,
                           desc=desc,
                           slug=patch_name,
                           post_data=post_data)

patches = troi.utils.discover_patches(PATCH_FOLDER)
for patch in patches:
    slug = patches[patch]().slug()
    app.add_url_rule('/%s' % slug, slug, web_patch_handler)



import sys
import urllib
from urllib.parse import quote

import requests
import ujson

from troi import Element, Release


class MSBMappingLookupElement(Element):
    '''
       Look up MBIDs for the given recordings, if possible. If recordings
       are not found, their data remains unchanged.
    '''

    SERVER_URL = "http://bono.metabrainz.org:8000/msid-mapping/json"

    def __init__(self, remove_unmatched=False):
        self.remove_unmatched = remove_unmatched

    @staticmethod
    def inputs():
        return [Recording]

    @staticmethod
    def outputs():
        return [Recording]

    def read(self, inputs, debug=False):

        in_recordings = inputs[0]
        artists = ",".join([ r.artist.name for r in in_recordings ])
        recordings = ",".join([ r.name for r in in_recordings ])

        url = self.SERVER_URL + "?[msb_artist_credit_name]=" + quote(artists) + \
            "&[msb_recording_name]=" + quote(recordings)

        r = requests.get(url)
        if r.status_code != 200:
            r.raise_for_status()

        try:
            mappings = ujson.loads(r.text)
        except Exception as err:
            raise RuntimeError(str(err))


        entities = []
        for row in mappings:
            r = in_recordings[int(row['index'])]

            if not row['mb_artist_name']:
                if not self.remove_unmatched:
                    entities.append(r)
                continue

            if r.mbid:
                r.add_note("recording mbid %s overwritten by msb_lookup" % (r.mbid))
            r.mbid = row['mb_recording_mbid']
            r.name = row['mb_recording_name']

            if r.artist.artist_credit_id:
                r.artist.add_note("artist_credit_id %d overwritten by msb_lookup" % (r.artist.artist_credit_id))
            r.artist.artist_credit_id = row['mb_artist_credit_id']
            r.artist.name = row['mb_artist_name']

            if r.release:
                if r.release.mbid:
                    r.release.add_note("mbid %d overwritten by msb_lookup" % (r.release.mbid))
                r.release.mbid = row['mb_release_mbid']
                r.release.name = row['mb_release_name']
            else:
                r.release = Release(row['mb_release_name'], mbid=row['mb_release_mbid'])

            entities.append(r)

        return entities



import requests
import ujson
from troi import Element, Artist, Release, Recording
import pylistenbrainz
import pylistenbrainz.errors

MAX_NUM_RECORDINGS_PER_REQUEST = 100


class UserRecordingRecommendationsElement(Element):
    '''
        Fetch recommended recording for a user from ListenBrainz
    '''

    def __init__(self, user_name, artist_type, count=25, offset=0):
        Element.__init__(self)
        self.client = pylistenbrainz.ListenBrainz()
        self.user_name = user_name
        self.count = count
        self.offset = offset
        self.artist_type = artist_type
        self._last_updated = None

    def outputs(self):
        return [Recording]

    @property
    def last_updated(self):
        return self._last_updated

    def read(self, inputs = [], debug=False):
        recording_list = []
        recordings = []

        while True:
            try:
                recordings = self.client.get_user_recommendation_recordings(self.user_name, 
                                                                            self.artist_type, 
                                                                            count=MAX_NUM_RECORDINGS_PER_REQUEST,
                                                                            offset=self.offset+len(recording_list))
            except (requests.exceptions.HTTPError, pylistenbrainz.errors.ListenBrainzAPIException) as err:
                if not str(err):
                    err = "Cannot fetch recommendations. Does the user '%s' exist?" % self.user_name
                raise RuntimeError(err)

            if not recordings or not len(recordings['payload']['mbids']):
                break

            for r in recordings['payload']['mbids']:
                recording_list.append(Recording(mbid=r['recording_mbid'], ranking=r['score']))

            if self.count > 0 and len(recording_list) >= self.count:
                break

        if recordings:
            self._last_updated = recordings['payload']['last_updated']

        return recording_list



import importlib
import inspect
import os
import traceback

from troi import Element, Artist, Release, Recording
import troi.patch


def discover_patches(patch_dir):

    patch_dict = {}
    for path in os.listdir(patch_dir):
        if path in ['.', '..']:
            continue

        if path.startswith("__init__"):
            continue

        if path.endswith(".py"):
            try:
                patch = importlib.import_module(os.path.basename(patch_dir) + "." + path[:-3])
            except ImportError as err:
                print("Cannot import %s, skipping:" % (path))
                traceback.print_exc()
                continue

            for member in inspect.getmembers(patch):
                if inspect.isclass(member[1]):
                    if issubclass(member[1], troi.patch.Patch):
                        patch_dict[member[1].slug()] = member[1]

    return patch_dict


def print_entity_list(entities, count=0):

    if len(entities) == 0:
        print("[ empty entity list ]")
        return

    if count == 0:
        count = len(entities)

    if isinstance(entities[0], Artist): 
        print("artist list")
        for e in entities[:count]:
            print("  %s %s" % (e.mbid)[:5], e.name)
    elif isinstance(entities[0], Recording): 
        print("recording list")
        for e in entities[:count]:
            if e.artist:
                print("  %s %-41s %s" % (e.mbid[:5], e.name[:80], e.artist.name[:60]))
            else:
                print("  %s %-41s" % (e.mbid[:5], e.name[:80] if e.name else ""))

    print()


class DumpElement(Element):
    """
        Accept whatever and print it out in a reasonably sane manner.
    """

    def __init__(self):
        pass

    @staticmethod
    def inputs():
        return []

    def read(self, inputs, debug=False):

        for input in inputs:
            print_entity_list(input)



import sys
import urllib
from urllib.parse import quote

import requests
import ujson

from troi import Element, Artist


class ArtistCreditIdLookupElement(Element):
    '''
       Look up MBIDs for the given artist_credit_ids
    '''

    SERVER_URL = "http://bono.metabrainz.org:8000/artist-credit-id-lookup/json"

    def __init__(self):
        pass

    @staticmethod
    def inputs(self):
        return [Artist]

    @staticmethod
    def outputs(self):
        return [Artist]

    def read(self, inputs, debug=False):

        ac_ids = []
        index = {}
        for a in inputs[0]:
            ac_ids.append(str(a.artist_credit_id))
            index[a.artist_credit_id] = a

        url = self.SERVER_URL + "?[artist_credit_id]=" + quote(",".join(ac_ids))

        r = requests.get(url)
        if r.status_code != 200:
            r.raise_for_status()

        try:
            artists = ujson.loads(r.text)
        except Exception as err:
            raise RuntimeError(str(err))

        entities = []
        for row in artists:
            a = index[row['artist_credit_id']]
            a.name = row['artist_credit_name']
            a.mbids = row['artist_credit_mbids']
            entities.append(a)

        return entities



from abc import ABC, abstractmethod

import troi


class Patch(ABC):

    def __init__(self):
        pass

    @staticmethod
    def inputs():
        '''
            This function should return a list of dicts that defined the type, name, description and if the 
            argument is optional. MusicBrainz entities and python base types can all be used. Example:
            [
                {
                    "type" : int,
                    "name": "num_recordings",
                    "optional" : True,
                    "desc" : "number of recorings"
                }
            ]
        '''
        return None

    @staticmethod
    def slug():
        ''' 
            Return the slug for this patch -- this is a URL friendly short identifier
            that can be used to invoke this patch via an HTTP call.

            e.g area-random-recordings
        '''

        return None

    @staticmethod
    def description():
        ''' 
            Return the description for this patch -- this short description (not more than a paragraph)
            should give the user an idea as to what the patch does.

            e.g "Generate a list of random recordings from a given area."
        '''

        return None

    @abstractmethod
    def create(input_args):
        '''
            The function creates the data pipeline and then returns it.
        '''

        return None



import datetime
import random

from troi import Element, Recording
import troi.listenbrainz.recs
import troi.filters
import troi.musicbrainz.recording_lookup


class DailyJamsElement(Element):
    '''
        Split weekly recommended recordings into 7 sets, one for each day of the week.
    '''

    def __init__(self, recs, day):
        Element.__init__(self)
        self.recs = recs
        self.day = day

    @staticmethod
    def inputs():
        return [Recording]

    @staticmethod
    def outputs():
        return [Recording]

    def read(self, inputs = [], debug=False):
        recordings = inputs[0]
        if not recordings or len(recordings) == 0:
            return []

        random.seed(self.recs.last_updated)
        random.shuffle(recordings)
        num_per_day = len(recordings) // 7
        days = [recordings[i:i + num_per_day] for i in range(0, len(recordings), num_per_day)]

        return days[self.day - 1]


class DailyJamsPatch(troi.patch.Patch):

    @staticmethod
    def inputs():
        return [{ "type": str, "name": "user_name", "desc": "MusicBrainz user name", "optional": False },
                { "type": str, "name": "type", "desc": "The type of daily jam. Must be 'top' or 'similar'.", "optional": False },
                { "type": int, "name": "day", "desc": "The day of the week to generate jams for (1-7). Leave blank for today.", "optional": True }]
        
    @staticmethod
    def outputs():
        return [Recording]

    @staticmethod
    def slug():
        return "daily-jams"

    @staticmethod
    def description():
        return "Generate a daily playlist from the ListenBrainz recommended recordings. Day 1 = Monday, Day 2  = Tuesday ..."

    def create(self, inputs):
        user_name = inputs[0]
        type = inputs[1]
        try:
            day = inputs[2]
            if day == None:
                day = 0
        except IndexError:
            day = 0

        if day > 7:
            raise RuntimeError("day must be an integer between 0-7.")
        if day == 0:
            day = datetime.datetime.today().weekday() + 1

        if type not in ("top", "similar"):
            raise RuntimeError("type must be either 'top' or 'similar'")


        recs = troi.listenbrainz.recs.UserRecordingRecommendationsElement(user_name=user_name,
                                                                          artist_type=type,
                                                                          count=-1)
        r_lookup = troi.musicbrainz.recording_lookup.RecordingLookupElement()
        r_lookup.set_sources(recs)

        # If an artist should never appear in a playlist, add the artist_credit_id here
        artist_filter = troi.filters.ArtistCreditFilterElement([])
        artist_filter.set_sources(r_lookup)

        artist_limiter = troi.filters.ArtistCreditLimiterElement()
        artist_limiter.set_sources(artist_filter)

        jams = DailyJamsElement(recs, day=day)
        jams.set_sources(artist_limiter)

        return jams



import datetime
import requests.exceptions

from troi import Recording
import troi.listenbrainz.area_random_recordings
import troi.tools.area_lookup
import troi.musicbrainz.recording_lookup
import troi.patch
import troi.filters


class AreaRandomRecordingsPatch(troi.patch.Patch):

    @staticmethod
    def inputs():
        return [{ "type": str, "name": "area", "desc": "MusicBrainz area from which to choose tracks.", "optional": False },
                { "type": int, "name": "start_year", "desc": "Start year", "optional": False },
                { "type": int, "name": "end_year", "desc": "End year", "optional": False }]

    @staticmethod
    def outputs():
        return [Recording]

    @staticmethod
    def slug():
        return "area-random-recordings"

    @staticmethod
    def description():
        return "Generate a list of random recordings from a given area."

    def create(self, inputs):
        area_name = inputs[0]
        start_year = inputs[1]
        end_year = inputs[2]

        area_id = troi.tools.area_lookup.area_lookup(area_name)

        if not start_year or start_year < 1800 or start_year > datetime.datetime.today().year:
            raise RuntimeError("start_year must be given and be an integer between 1800 and the current year.")

        if not end_year or end_year < 1800 or end_year > datetime.datetime.today().year:
            raise RuntimeError("end_year must be given and be an integer between 1800 and the current year.")

        if end_year < start_year:
            raise RuntimeError("end_year must be equal to or greater than start_year.")

        area = troi.listenbrainz.area_random_recordings.AreaRandomTracksElement(area_id, start_year, end_year)
        r_lookup = troi.musicbrainz.recording_lookup.RecordingLookupElement()
        r_lookup.set_sources(area)

        artist_limiter = troi.filters.ArtistCreditLimiterElement()
        artist_limiter.set_sources(r_lookup)

        return artist_limiter



from collections import defaultdict
from operator import itemgetter
from random import shuffle

import troi


class ArtistCreditFilterElement(troi.Element):
    '''
        Remove recordings if they do or do not belong to a given list of artists.
    '''

    def __init__(self, artist_credit_ids, include=False):
        '''
            Filter a list of Recordings based on their artist_credit_id.
            The default behaviour is to exclude the artists given, but if
            include=True then only the artists in the given list pass
            the filter. Throws RuntimeError is not all recordings have
            artist_credit_ids set.
        '''
        self.artist_credit_ids = artist_credit_ids
        self.include = include

    @staticmethod
    def inputs():
        return [Recording]

    @staticmethod
    def outputs():
        return [Recording]

    def read(self, inputs, debug=False):

        recordings = inputs[0]

        ac_index = {}
        for ac in self.artist_credit_ids:
            try:
                ac_index[ac] = 1
            except KeyError:
                raise RuntimeError(self.__name__ + " needs to have all input recordings to have artist.artist_credit_id defined!")

        results = []
        for r in recordings:
            if not r.artist or not r.artist.artist_credit_id:
                if debug:
                    print("- debug recording %s has not artist credit id" % (r.mbid))
                continue

            if self.include:
                if r.artist.artist_credit_id in ac_index:
                    results.append(r)
            else:
                if r.artist.artist_credit_id not in ac_index:
                    results.append(r)

        return results


class ArtistCreditLimiterElement(troi.Element):

    def __init__(self, count=2, exclude_lower_ranked=True):
        '''
            This element examines there passed in recordings and if the count of
            recordings by any one artists exceeds the given limit, excessive recordigns
            are removed. If the flag exclude_lower_ranked is True, then the lowest
            ranked recordings are removed, otherwise the highest ranked recordings
            are removed. Throws RuntimeError is not all recordings have
            artist_credit_ids set.
        '''
        self.count = count
        self.exclude_lower_ranked = exclude_lower_ranked

    @staticmethod
    def inputs():
        return [Recording]

    @staticmethod
    def outputs():
        return [Recording]

    def read(self, inputs, debug=False):

        recordings = inputs[0]
        ac_index = defaultdict(list)
        all_have_rankings = True
        for rec in recordings:
            try:
                ac_index[rec.artist.artist_credit_id].append((rec.mbid, rec.ranking))
                if rec.ranking == None:
                    all_have_rankings = False
            except KeyError:
                raise RuntimeError(self.__name__ + " needs to have all input recordings to have artist.artist_credit_id defined!")

        for key in ac_index:
            if all_have_rankings:
                ac_index[key] = sorted(ac_index[key], key=itemgetter(1), reverse=self.exclude_lower_ranked)
            else:
                shuffle(ac_index[key])
            ac_index[key] = ac_index[key][:self.count]

        pass_recs = []
        for key in ac_index:
            for mbid, ranking in ac_index[key]:
                pass_recs.append(mbid)

        results = []
        for r in recordings:
            if r.mbid in pass_recs:
                results.append(r)

        return results



import copy
import troi

def is_homogeneous(entities):
    '''
        Check to see if all the items in a list of of the same type
    '''

    if not entities:
        return True

    type_set = set()
    for e in entities:
        type_set.add(type(e))

    return len(type_set) == 1


def _ensure_conformity(entities_0, entities_1):
    '''
        Check that both entity lists are homogenous and of the same type.
    '''

    if not is_homogeneous(entities_0):
        raise TypeError("entities_0 list not homogenous")

    if not is_homogeneous(entities_1):
        raise TypeError("entities_1 list not homogenous")

    if entities_0 and entities_1 and type(entities_0[0]) != type(entities_1[0]):
        raise TypeError("entities_0 and entities_1 must both be homogenous with all the same type")

    return True


class UniqueElement(troi.Element):
    '''
        Make this passed list of entities unique base on the passed key
        (must be one of name, mbid or msid) and return the unique list.
        Currently the order of the list is not preserved. This should be improved on...
    '''

    def __init__(self, key = "mbid"):
        self.key = key

    @staticmethod
    def inputs():
        return []

    def read(self, entities_arg, debug=False):
        entities = entities_arg[0]

        if not entities:
            return []

        if not is_homogeneous(entities):
            raise TypeError("entity list not homogenous")

        if isinstance(entities[0], troi.Artist):
            if self.key not in ['mbids', 'msid', 'name', 'artist_credit_id']:
                raise ValueError("key must be one of mbid/s, msid, name or artist_credit_id.")
        else:
            if self.key not in ['mbid', 'msid', 'name']:
                raise ValueError("key must be one of mbid/s, msid or name.")

        entity_dict = {}
        for e in entities:
            if isinstance(e, troi.Artist) and self.key == "mbids":
                entity_dict[",".join(getattr(e, self.key))] = e
            else:
                entity_dict[getattr(e, self.key)] = e

        return list(entity_dict.values())


class UnionElement(troi.Element):
    '''
        Combine both entities lists into one
    '''

    def __init__(self):
        pass

    def inputs(self):
        return []

    def read(self, entities, debug=False):
        entities_0 = entities[0]
        entities_1 = entities[1]

        if not entities_0:
            return entities_1

        if not entities_1:
            return entities_0

        _ensure_conformity(entities_0, entities_1)

        result = copy.copy(entities_0)
        result.extend(entities_1)

        return result


class IntersectionElement(troi.Element):
    """
        Return the list of entities that exist in both entities lists.
    """

    def __init__(self, key = "mbid"):
        self.key = key

    def inputs(self):
        return []

    def read(self, entities, debug=False):
        entities_0 = entities[0]
        entities_1 = entities[1]

        if not entities_0 or not entities_1:
            return []

        _ensure_conformity(entities_0, entities_1)

        if isinstance(entities_0[0], troi.Artist):
            if self.key not in ['mbids', 'msid', 'name', 'artist_credit_id']:
                raise ValueError("key must be one of mbid/s, msid, name or artist_credit_id.")
        else:
            if self.key not in ['mbid', 'msid', 'name']:
                raise ValueError("key must be one of mbid/s, msid or name.")

        entity_dict = {}
        for e in entities_1:
            if isinstance(e, troi.Artist) and self.key == "mbids":
                entity_dict[",".join(getattr(e, self.key))] = e
            else:
                entity_dict[getattr(e, self.key)] = e

        results = []
        for e in entities_0:
            if isinstance(e, troi.Artist) and self.key == "mbids":
                if ",".join(getattr(e, self.key)) in entity_dict:
                    results.append(e)
            else:
                if getattr(e, self.key) in entity_dict:
                    results.append(e)


        return results


class DifferenceElement(troi.Element):
    '''
        Return the list of recordings in entities_0 minus those in entities_1
    '''

    def __init__(self, key = "mbid"):
        self.key = key

    def inputs(self):
        return []

    def read(self, entities, debug=False):
        entities_0 = entities[0]
        entities_1 = entities[1]

        if not entities_0:
            return []

        if not entities_1:
            return entities_0

        _ensure_conformity(entities_0, entities_1)

        if isinstance(entities_0[0], troi.Artist):
            if self.key not in ['mbids', 'msid', 'name', 'artist_credit_id']:
                raise ValueError("key must be one of mbid/s, msid, name or artist_credit_id.")
        else:
            if self.key not in ['mbid', 'msid', 'name']:
                raise ValueError("key must be one of mbid/s, msid or name.")

        entity_dict = {}
        for e in entities_1:
            if isinstance(e, troi.Artist) and self.key == "mbids":
                entity_dict[",".join(getattr(e, self.key))] = e
            else:
                entity_dict[getattr(e, self.key)] = e

        results = []
        for e in entities_0:
            if isinstance(e, troi.Artist) and self.key == "mbids":
                if ",".join(getattr(e, self.key)) not in entity_dict:
                    results.append(e)
            else:
                if getattr(e, self.key) not in entity_dict:
                    results.append(e)

        return results



import sys
import urllib
from urllib.parse import quote

import requests
import ujson

from troi import Element, Recording


class AnnoyLookupElement(Element):
    '''
        Given an recording MBID, lookup tracks that are similar given some 
        criteria (e.g. mfccs, gfccs, etc).
    '''

    SERVER_URL = "http://similarity.acousticbrainz.org/api/v1/similarity/"

    def __init__(self, metric, mbid):
        '''
            The given recording mbid is the source track that will be looked up 
            in the annoy index using the passed metric.
        '''
        Element.__init__(self)
        self.mbid = mbid
        self.metric = metric

    def outputs(self):
        return [Recording]

    def read(self, inputs, debug=False):
        print("  annoy: read for %s/%s" % (self.metric, self.mbid))

        url = self.SERVER_URL + self.metric + "/" + self.mbid
        r = requests.get(url)
        if r.status_code != 200:
            r.raise_for_status()

        try:
            results = ujson.loads(r.text)
        except Exception as err:
            raise RuntimeError(str(err))


        entities = []
        for row in results:
            r = Recording(mbid=row['recording_mbid'], 
                          acousticbrainz={
                              'similarity': row['distance'], 
                              'offset' : row['offset']
                          }
                         )
            r.add_note("Related to %s" % self.mbid)
            entities.append(r)

        print("  annoy: read %d recordings" % len(entities))

        return entities



from setuptools import setup, find_packages

with open("README.md", "r") as fh:
    long_description = fh.read()

setup(
    name="troi",
    version="0.0.1pre",
    author="MetaBrainz Foundation",
    description="An empathetic music recommendation system pipeline",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/metabrainz/troi-recommendation-playground",
    install_requires=["click>=7", "ujson>=2", "requests>=2.24", "pylistenbrainz", "openpost"],
    packages=find_packages(),
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    python_requires='>=3.8',
)



import sys
import uuid
from urllib.parse import quote

import requests
import ujson

from troi import Element, Artist, Recording

class AreaRandomTracksElement(Element):
    '''
        Given an area, find random tracks from releases that were released
        by artists from that given area.
    '''

    SERVER_URL = "http://bono.metabrainz.org:8000/area-random-recordings/json"

    def __init__(self, area_id, start_year=0, end_year=3000):
        Element.__init__(self)
        self.area_id = area_id
        self.start_year = start_year
        self.end_year = end_year

    @staticmethod
    def inputs():
        return [ ]

    @staticmethod
    def outputs():
        return [ Recording ]

    def read(self, inputs, debug=False):

        data = [ { 'area_id': self.area_id, 'start_year' : self.start_year, 'end_year' : self.end_year } ]
        r = requests.post(self.SERVER_URL, json=data)
        if r.status_code != 200:
            r.raise_for_status()

        try:
            rows = ujson.loads(r.text)
        except Exception as err:
            raise RuntimeError(str(err))

        recordings = []
        for row in rows:
            recordings.append(Recording(mbid=row['recording_mbid'], 
                                        name=row['recording_name'], 
                                        artist=Artist(name=row['artist_credit_name'],
                                                      artist_credit_id=row['artist_credit_id'])))

        return recordings



import sys
import urllib
import copy
from collections import defaultdict
from urllib.parse import quote

import requests
import ujson

from troi import Element, Release


class RelatedArtistCreditsElement(Element):
    '''
        Look up related artist_credits, given a list of artists_credits
    '''

    SERVER_URL = "http://bono.metabrainz.org:8000/artist-credit-similarity/json"

    def __init__(self, threshold=0):
        self.threshold = threshold

    @staticmethod
    def inputs():
        return [Recording]

    @staticmethod
    def outputs():
        return [Recording]

    def read(self, inputs, debug=False):

        artists = inputs[0]
        ac_ids = ",".join([ str(a.artist_credit_id) for a in artists ])
        url = self.SERVER_URL + "?[artist_credit_id]=" + quote(ac_ids) + \
            "&threshold=%d" % self.threshold

        r = requests.get(url)
        if r.status_code != 200:
            r.raise_for_status()

        try:
            relations = ujson.loads(r.text)
        except Exception as err:
            raise RuntimeError(str(err))

        index = defaultdict(list)
        for row in relations:
            index[row['artist_credit_id']].append(row)

        entities = []
        for artist in artists:
            a = copy.deepcopy(artist)
            a.mb['related_artist_credit_ids'] = index[artist.artist_credit_id]
            entities.append(a)

        return entities



from abc import ABC, abstractmethod


class Element(ABC):
    """
        Base class for elements
    """

    def __init__(self):
        self.sources = []

    def set_sources(self, sources):
        """
        """

        # TODO: add type checking of the pipeline
        if not isinstance(sources, list):
            sources = [ sources ]

        self.sources = sources

    def check(self):
        """
            Check to see if the necessary connections are in place
            and that types match correctly.
        """

        if not self.sources:
            raise RuntimeError("element %s has no sources defined." % str(type(self)))

        for source in self.sources:
            source.check()


    def generate(self, debug=False):
        """
            Generate output from the pipeline. This should be called on
            the last element in the pipeline and no where else. At the root
            node generate returns the results, but on interior nodes it
            returns data to be used as input in the next level.
        """

        source_lists = []
        if self.sources:
            for source in self.sources:
                source_lists.append(source.generate(debug))
           
        recordings = self.read(source_lists, debug)
        if debug:
            print("- debug %-50s %d items" % (type(self).__name__[:49], len(recordings or [])))

        return recordings
      
    def run(self):
        """ 
            This function should be called on the very last element of the
            pipeline to generate output from the timeline. Run will santiy
            check the pipeline first, and then generate output if it can.
        """
        self.check()
        return self.generate()
            
    @staticmethod
    def inputs():
        """
            Return a list of Artist, Release or Recording classes that define
            the type and number of input lists to this element. 
            e.g. [ Artist, Recording ] means that this element expects
            a list of artists and a list of recordings for inputs.
        """
        return None

    @staticmethod
    def outputs():
        """
            Return a list of Artist, Release or Recording classes that define
            the type and number of output lists returned by this element. 
            e.g. [ Artist, Recording ] means that this element returns
            a list of artists and a list of recordings.
        """
        return None

    @abstractmethod
    def read(self, source_data_list, debug):
        ''' 
            This method is where the action happens -- when the consumer wants to 
            read data from the pipeline, it calls read() on the last element in
            the pipeline and this casues the while pipeline to generate result.
            If the initializers of other objects in the pipeline are updated,
            calling read() again will generate the set new. Passing True for
            debug should print helpful debug statements about its progress.
        '''

        pass


class Entity(ABC):
    """
        This is the base class for entity objects in troi. Each object will have
        a name, mbid, msid attributes which are all optional. There will also be
        three dicts named musicbrainz, listenbrainz and acousticbrainz that will
        contain collected metadata respective of each project. 

        For instance, the musicbrainz dict might have keys that shows the type
        of an artist or the listenbrainz dict might contain the BPM for a track.
        How exactly these dicts will be organized is TDB.
    """
    def __init__(self, ranking=None, musicbrainz={}, listenbrainz={}, acousticbrainz={}):
        self.name = None
        self.mbid = None
        self.msid = None
        self.musicbrainz = musicbrainz
        self.listenbrainz = listenbrainz
        self.acousticbrainz = acousticbrainz
        self.notes = []
        self.ranking = ranking

    @property
    def mb(self):
        return self.musicbrainz

    @property
    def lb(self):
        return self.listenbrainz

    @property
    def ab(self):
        return self.acousticbrainz

    def add_note(self, note):
        '''
            This allows various parts of the pipeline to leave notes/comments in the data/
        '''
        self.notes.append(note)

    def __str__(self):
        return "<Entity()>"


class Area(Entity):
    """
        The class that represents an area.
    """
    def __init__(self, id=id, name=None):
        Entity.__init__(self)
        self.name = name
        self.id = id

    def __str__(self):
        return "<Area('%s', %d)>" % (self.name, self.id)


class Artist(Entity):
    """
        The class that represents an artist.
    """
    def __init__(self, name=None, ranking=None, mbids=None, msid=None, artist_credit_id=None,
                 musicbrainz={}, listenbrainz={}, acousticbrainz={}):
        Entity.__init__(self, ranking, musicbrainz, listenbrainz, acousticbrainz)
        self.name = name
        self.artist_credit_id = artist_credit_id
        if mbids:
            if not isinstance(mbids, list):
                raise TypeError("Artist mbids must be a list.")
            self.mbids = sorted(mbids)
        else:
            self.mbids = None
        self.msid = msid

    def __str__(self):
        return "<Artist('%s', [%s], %s, %s)>" % (self.name, ",".join(self.mbids or []), self.msid, self.artist_credit_id)


class Release(Entity):
    """
        The class that represents a release.
    """
    def __init__(self, name=None, ranking=None, mbid=None, msid=None, artist=None, 
                  musicbrainz={}, listenbrainz={}, acousticbrainz={}):
        Entity.__init__(self, ranking, musicbrainz, listenbrainz, acousticbrainz)
        self.artist = artist
        self.name = name
        self.mbid = mbid
        self.msid = msid

    def __str__(self):
        return "<Release('%s', %s, %s)>" % (self.name, self.mbid, self.msid)


class Recording(Entity):
    """
        The class that represents a recording.
    """
    def __init__(self, name=None, ranking=None, mbid=None, msid=None, length=None, artist=None, release=None, 
                  musicbrainz={}, listenbrainz={}, acousticbrainz={}):
        Entity.__init__(self, ranking, musicbrainz, listenbrainz, acousticbrainz)
        self.length = length # track length in ms
        self.artist = artist
        self.release = release
        self.name = name
        self.mbid = mbid
        self.msid = msid

    def __str__(self):
        return "<Recording('%s', %s, %s)>" % (self.name, self.mbid, self.msid)



import sys
import uuid
from urllib.parse import quote

import requests
import ujson

from troi import Element, Artist, Recording

class RecordingLookupElement(Element):
    '''
        Look up a musicbrainz data for a list of recordings, based on MBID. 
    '''

    SERVER_URL = "https://labs.api.listenbrainz.org/recording-mbid-lookup/json?count=%d"

    @staticmethod
    def inputs():
        return [ Recording ]

    @staticmethod
    def outputs():
        return [ Recording ]

    def read(self, inputs, debug=False):

        recordings = inputs[0]
        if not recordings:
            return []

        data = []
        r_mbids = ",".join([ r.mbid for r in recordings ])
        for r in recordings:
            data.append({ '[recording_mbid]': r.mbid })

        if debug:
            print("- debug %d recordings" % len(recordings))

        r = requests.post(self.SERVER_URL % len(recordings), json=data)
        if r.status_code != 200:
            r.raise_for_status()

        try:
            rows = ujson.loads(r.text)
            if debug:
                print("- debug %d rows in response" % len(rows))
        except Exception as err:
            raise RuntimeError(str(err))

        mbid_index = {}
        for row in rows:
            mbid_index[row['original_recording_mbid']] = row

        for r in recordings:
            try:
                row = mbid_index[r.mbid]
            except KeyError:
                if debug:
                    print("- debug recording MBID %s not found, skipping." % r.mbid)
                continue

            if not r.artist:
                a = Artist(name=row['artist_credit_name'],
                           mbids=row.get('[artist_credit_mbids]', []),
                           artist_credit_id=row['artist_credit_id'])
                r.artist = a
            else:
                r.artist.name = row['artist_credit_name']
                r.artist.mbids = row.get('[artist_credit_mbids]', []),
                r.artist.artist_credit_id = row['artist_credit_id']

            r.name = row['recording_name']
            r.length = row['length']

        return recordings
